import logging
from time import perf_counter

import numpy as np


class Timer:
    def __enter__(self):
        self.start = perf_counter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.elapsed_time = perf_counter() - self.start


class DataHandler:
    def __init__(self, var, t0, cell_slice):
        self.var = var
        self.cell_slice = cell_slice
        self.loco_server = None
        mi, ma, st = cell_slice.indices(var.shape[-1])
        assert st == 1, "non-contiguous cell slices are not supported"
        self.t = t0
        self.prune_offset = var.attrs.get("loco::num_retained_timesteps", None)
        self.t_flushed = self.t
        self.time_chunk_size = var.chunks[0]
        self.use_buffer = var.attrs.get("hiopy::use_buffer", self.time_chunk_size > 1)
        if self.use_buffer:
            self.buf = np.empty([self.time_chunk_size, *var.shape[1:-1], ma - mi], dtype=var.dtype)

    def flush(self):
        if self.prune_offset is not None:
            fro = max(0, self.t_flushed - self.prune_offset)
            to = max(0, self.t - self.prune_offset)
            with Timer() as t:
                self.var[fro:to, ..., self.cell_slice] = self.var.fill_value
            logging.info(f"pruned {self.var.name} for timesteps {fro}:{to}. Took {t.elapsed_time}s")

        fro = self.t_flushed % self.time_chunk_size
        to = fro + (self.t - self.t_flushed)
        with Timer() as t:
            self.var[self.t_flushed : self.t, ..., self.cell_slice] = self.buf[fro:to, ...]
        logging.info(
            f"wrote {self.var.name} for timesteps {self.t_flushed}:{self.t}."
            f" Took {t.elapsed_time}s"
        )
        self.t_flushed = self.t
        if self.loco_server is not None:
            self.loco_server.set_timestep(self.var.name, self.t - 1)

    def __call__(self, event):
        logging.info(
            f"received {self.var.name} at {self.t % self.time_chunk_size+1}/{self.time_chunk_size}"
        )
        if self.use_buffer:
            self.buf[self.t % self.time_chunk_size, ...] = np.transpose(event.data)
            event.release()
        else:
            self.buf = np.squeeze(np.transpose(event.data))[None, ...]
        self.t += 1
        if self.t % self.time_chunk_size == 0 or not self.use_buffer:  # chunk complete
            self.flush()

    def set_loco_server(self, loco_server):
        self.loco_server = loco_server
