import numpy as np

from ._zarr_utils import get_var_group


# computes an id that identifies the grid of a variable
def grid_id(var):
    """
    var: dtype zarr.array
    zgroup: dtype zarr.group
    """
    assert "grid_mapping" in var.attrs

    zgroup = get_var_group(var)
    crs = zgroup[var.attrs["grid_mapping"]]

    var_group = var.attrs.get("hiopy::var_group", "")
    spatial_chunk_shape = var.chunks[-1]

    # healpix
    if "grid_mapping_name" in crs.attrs and crs.attrs["grid_mapping_name"] == "healpix":
        s = var_group + "healpix_"
        s += str(crs.attrs["healpix_nside"])
        s += str(crs.attrs["healpix_order"])
        s += "_" + str(spatial_chunk_shape)
        # if the cell dimension has a coordinate, take it into account
        if "_ARRAY_DIMENSIONS" in var.attrs and var.attrs["_ARRAY_DIMENSIONS"][-1] in zgroup:
            s += hash(str(zgroup[var.attrs["_ARRAY_DIMENSIONS"][-1]][:]))
        return s

    # Meteogram
    if "coordinates" in crs.attrs:
        try:
            lon, lat = crs.attrs["coordinates"].split(" ")
            s = hash((str(zgroup[lat][:]), str(zgroup[lon][:]), var_group, spatial_chunk_shape))
            return s
        except Exception as e:
            raise RuntimeError("Cannot parse coordinates") from e

    raise RuntimeError("Unknown grid")


def def_grid(coyote, var, chunk_slice):
    zgroup = get_var_group(var)
    crs = zgroup[var.attrs["grid_mapping"]]
    # healpix
    if "grid_mapping_name" in crs.attrs and crs.attrs["grid_mapping_name"] == "healpix":
        nside = crs.attrs["healpix_nside"]
        if "_ARRAY_DIMENSIONS" in var.attrs and var.attrs["_ARRAY_DIMENSIONS"][-1] in zgroup:
            cellidx = zgroup[var.attrs["_ARRAY_DIMENSIONS"][-1]][chunk_slice]
        else:
            ncells_global = 12 * (nside**2)
            mi, ma, st = chunk_slice.indices(ncells_global)
            cellidx = np.arange(mi, ma, st, dtype=np.intc)
        def_healpix_grid(coyote, nside, nest=True, cell_idx=cellidx)
        return

    # Meteogram
    if "coordinates" in crs.attrs:
        lon, lat = crs.attrs["coordinates"].split(" ")
        rad_multiplier = np.pi / 180
        coyote.def_grid([], rad_multiplier * zgroup[lon][:], rad_multiplier * zgroup[lat][:], [])
        return

    raise RuntimeError(f"Unknown grid in {var.path}")


def _xyz2lonlat(xyz):
    xyz = np.array(xyz)
    lat = np.arcsin(xyz[..., 2])
    lon = np.arctan2(xyz[..., 1], xyz[..., 0])
    return lon, lat


def def_healpix_grid(coyote, nside, nest=True, cell_idx=None):
    import healpy

    if cell_idx is None:
        ncells = healpy.pixelfunc.nside2npix(nside)
        cell_idx = range(ncells)

    centers_xyz = np.stack(
        healpy.pixelfunc.pix2vec(nside, cell_idx, nest=nest),
        axis=-1,
    )
    clon, clat = _xyz2lonlat(centers_xyz)
    boundaries_xyz = healpy.boundaries(nside, cell_idx, nest=nest).transpose(0, 2, 1).reshape(-1, 3)
    verts_xyz, quads = np.unique(boundaries_xyz, return_inverse=True, axis=0)
    vlon, vlat = _xyz2lonlat(verts_xyz)
    vertex_of_cell = quads.reshape(-1, 4).T
    coyote.def_grid(4 * np.ones(len(cell_idx)), vlon, vlat, vertex_of_cell.T.flatten(), clon, clat)
