#!/usr/bin/env python3

import healpy
import numpy as np


def add_time(dataset, startdate, enddate, dt, name="time"):
    if type(startdate) is str:
        startdate = np.datetime64(startdate)
    if type(enddate) is str:
        enddate = np.datetime64(enddate)
    if type(dt) is int:
        dt = np.timedelta64(dt, "s")

    time_data = (np.arange(startdate + dt, enddate + dt, dt) - startdate) // np.timedelta64(1, "s")

    for g in _collect_groups(dataset):
        time = g.create_dataset(name, data=time_data, fill_value=None, shape=time_data.shape)
        time.attrs["_ARRAY_DIMENSIONS"] = (name,)
        time.attrs["axis"] = "T"
        time.attrs["calendar"] = "proleptic_gregorian"
        time.attrs["units"] = f"seconds since {startdate}"


def _collect_groups(dataset):
    if "crs" in dataset.array_keys():
        yield dataset
    for _name, g in dataset.groups():
        yield from _collect_groups(g)


def add_height(dataset, name, positive_direction: str, values: list[int]):
    for g in _collect_groups(dataset):
        height = g.create_array(name, fill_value=None, dtype=np.int64, shape=(len(values),))
        height[:] = np.asarray(values, dtype=np.int64)
        height.attrs["_ARRAY_DIMENSIONS"] = [name]
        height.attrs["axis"] = "Z"
        height.attrs["long_name"] = "generalized_height"
        height.attrs["positive"] = positive_direction
        height.attrs["standard_name"] = "height"


def add_variable(
    dataset,
    name,
    yac_source_comp,
    yac_source_grid,
    zaxis=None,
    taxis="time",
    time_method=None,
    chunk_shape=None,
    frac_mask=None,
    yac_name=None,
    attributes=None,
    chunks_per_shard=None,
    **kwargs,
):
    for g in _collect_groups(dataset):
        taxis_tuple = tuple() if taxis is None else (taxis,)
        ntime = tuple() if taxis is None else (g[taxis].shape[0],)
        grid_mapping_name = g["crs"].attrs["grid_mapping_name"]
        spatial_attr = "point" if (grid_mapping_name == "point_cloud") else "cell"
        crs_len = 0
        if grid_mapping_name == "healpix":
            crs_len = healpy.nside2npix(g["crs"].attrs["healpix_nside"])
        elif grid_mapping_name == "point_cloud":
            lon_coord, lat_coord = g["crs"].attrs["coordinates"].split(" ")
            assert lon_coord in g and lat_coord in g
            assert g[lon_coord].shape[0] == g[lat_coord].shape[0]
            crs_len = g[lat_coord].shape[0]
        else:
            raise Exception("Unknown crs.")

        _attributes = attributes or {}

        if zaxis is None:
            shape = (*ntime, crs_len)
            _attributes["_ARRAY_DIMENSIONS"] = (*taxis_tuple, spatial_attr)
        else:
            nheight = g[zaxis].shape[0]
            shape = (*ntime, nheight, crs_len)
            _attributes["_ARRAY_DIMENSIONS"] = (*taxis_tuple, zaxis, spatial_attr)

        _attributes["grid_mapping"] = "crs"
        _chunk_shape = "auto"
        _shard_shape = None

        if chunk_shape is not None:
            _chunk_shape = tuple(min(chunk_shape, shape))
            if chunks_per_shard is not None:
                _shard_shape = tuple(i * chunks_per_shard for i in _chunk_shape)

        v = g.create_array(
            name,
            shape=shape,
            dtype=np.float32,
            fill_value=np.nan,
            chunks=_chunk_shape,
            shards=_shard_shape,
            **kwargs,
        )

        # TODO: Use a generic name instead of hiopy such that it represents arbitrary grid too
        for key, val in _attributes.items():
            v.attrs[key] = val
        if time_method is not None:
            v.attrs["hiopy::time_method"] = time_method
        if "hiopy::parent" not in g.attrs:
            v.attrs["hiopy::yac_source"] = (yac_source_comp, yac_source_grid)
            if yac_name is not None:
                v.attrs["hiopy::src_name"] = yac_name
        else:
            v.attrs["hiopy::nnn"] = 4
        v.attrs["hiopy::enable"] = True


def extend_time(dataset, enddate):
    for g in _collect_groups(dataset):
        assert (
            "seconds since " in g.time.attrs["units"]
        ), "Currently the time must be given in seconds"
        refdate = np.datetime64(g.time.attrs["units"][len("seconds since ") :])
        dt = g.time[1] - g.time[0]
        enddate_s = (enddate - refdate) // np.timedelta64(1, "s")
        g.time.append(np.arange(g.time[-1] + dt, enddate_s + dt, dt))
        print(g)
