#include "field.hpp"

namespace coyote {

    Field::Field(const int& target_field_id,
                 const std::optional<int>& source_field_id,
                 const std::vector<double>& buffer,
                 const std::vector<double>& frac_mask,
                 const int& collection_size,
                 const std::function<void(Event)>& eventHandle)
        : _target_field_id(target_field_id),
          _source_field_id(source_field_id),
          _buffer(buffer),
          _frac_mask(frac_mask),
          _collection_size(collection_size),
          op(eventHandle) {
        _point_size = buffer.size() / collection_size;
    }

    void Field::put_forward() {
        if (_source_field_id) {
            // if the previous put is not completed yet, we need to
            // wait, because it might be that the source and target
            // field of the same exchange is on this process. In this
            // case YAC would trigger an abort (to avoid an deadlock in
            // the non-threaded case) if the put is not already
            // completed. This case should appear rarely in real-world
            // applications.
            int field_id = _source_field_id.value();
            int info     = 0;
            int flag;
            do {
                yac_ctest(field_id, &flag);
            } while (!flag);
            int ierror = 0;
            if (!_frac_mask.empty()) {
                yac_cput_frac_(
                    field_id, _collection_size, _buffer.data(), _frac_mask.data(), &info, &ierror);
            } else {
                yac_cput_(field_id, _collection_size, _buffer.data(), &info, &ierror);
            }
        }
    }

    std::string Field::datetime() const { return _datetime; }

    bool Field::is_done() const { return _info == YAC_ACTION_OUT_OF_BOUND; }
    bool Field::is_ready() {
        if (!_get_started) {
            if (!semaphore.try_acquire())
                return false;
            int ierror = 0;
            _datetime  = yac_cget_field_datetime(_target_field_id);
            yac_cget_async_(_target_field_id, _collection_size, _buffer.data(), &_info, &ierror);
            _get_started = true;
        }
        int ready = 0;
        yac_ctest(_target_field_id, &ready);
        if (!ready)
            return false;
        yac_cget_action(_target_field_id, &_info);
        _get_started = false;
        _datetime    = yac_cget_field_datetime(_target_field_id);
        return true;
    }

} // namespace coyote
