#include <ctype.h>
#include <math.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <mpi.h>
#include <yac.h>

void print_help() {
    fprintf(stderr, "simple_source.c [-e enddate] [-r resolution]\n\n");
    yac_abort(MPI_COMM_WORLD, "", __FILE__, __LINE__);
}

void parse_arguments(int argc, char** argv, char** enddate, int* resolution) {
    int opt;
    while ((opt = getopt(argc, argv, "he:r:f")) != -1) {
        switch (opt) {
        case 'h':
            print_help();
            break;
        case 'e':
            *enddate = optarg;
            break;
        case 'r':
            *resolution = atoi(optarg);
            break;
        case '?':
            if (isprint(optopt))
                fprintf(stderr, "Unknown option `-%c'.\n", optopt);
            else
                fprintf(stderr, "Unknown option character `\\x%x'.\n", optopt);
            yac_abort(MPI_COMM_WORLD, "Unknown command-line argument.", __FILE__, __LINE__);
        }
    }
}

int main(int argc, char** argv) {
    yac_cinit();
    yac_cdef_calendar(YAC_PROLEPTIC_GREGORIAN);
    char* enddate  = "2000-01-01T00:10:00.00Z";
    int resolution = 90;
    parse_arguments(argc, argv, &enddate, &resolution);
    yac_cdef_datetime("2000-01-01T00:00:00.00Z", enddate);
    printf("enddate: %s\n", enddate);
    int comp_id, grid_id, point_id;
    const char* comp_name = "simple_source";
    yac_cdef_comp(comp_name, &comp_id);
    int nbr_points[] = {2 * resolution, resolution};
    int cyclic[]     = {1, 0};
    double lon[nbr_points[0]];
    for (size_t i = 0; i < nbr_points[0]; ++i) {
        lon[i] = i * 2 * M_PI / nbr_points[0];
    }
    double lat[nbr_points[1]];
    for (size_t i = 0; i < nbr_points[1]; ++i) {
        lat[i] = -M_PI_2 + (i + 0.5) * M_PI / nbr_points[1];
    }
    const char* grid_name = "simple_grid";
    yac_cdef_grid_reg2d(grid_name, nbr_points, cyclic, lon, lat, &grid_id);
    int glb_idx[nbr_points[0] * nbr_points[1]];
    for (int i = 0; i < nbr_points[0] * nbr_points[1]; ++i)
        glb_idx[i] = i;
    yac_cset_global_index(glb_idx, YAC_LOCATION_CORNER, grid_id);
    yac_cdef_points_reg2d(grid_id, nbr_points, YAC_LOCATION_CORNER, lon, lat, &point_id);

    int x_field_id = -1;
    yac_cdef_field("x", comp_id, &point_id, 1, 1, "PT1S", YAC_TIME_UNIT_ISO_FORMAT, &x_field_id);
    yac_cdef_field_metadata(
        comp_name,
        grid_name,
        "x",
        "{\n"
        "  \"cf\": {\n"
        "    \"standard_name\": \"x\",\n"
        "    \"long_name\": \"this is a very long x\",\n"
        "    \"units\": \"1\"\n"
        "  }\n"
        "}");
    int y_field_id = -1;
    yac_cdef_field("y", comp_id, &point_id, 1, 1, "PT1S", YAC_TIME_UNIT_ISO_FORMAT, &y_field_id);
    yac_cdef_field_metadata(
        comp_name,
        grid_name,
        "y",
        "{\n"
        "  \"cf\": {\n"
        "    \"standard_name\": \"y\",\n"
        "    \"long_name\": \"this is a very long y\",\n"
        "    \"units\": \"1\"\n"
        "  }\n"
        "}");
    int z_field_id = -1;
    yac_cdef_field("z", comp_id, &point_id, 1, 1, "PT1S", YAC_TIME_UNIT_ISO_FORMAT, &z_field_id);
    yac_cdef_field_metadata(
        comp_name,
        grid_name,
        "z",
        "{\n"
        "  \"cf\": {\n"
        "    \"standard_name\": \"z\",\n"
        "    \"long_name\": \"this is a very long z\",\n"
        "    \"units\": \"1\"\n"
        "  }\n"
        "}");
    int clock_field_id = -1;
    yac_cdef_field(
        "clock", comp_id, &point_id, 1, 1, "PT1S", YAC_TIME_UNIT_ISO_FORMAT, &clock_field_id);
    yac_cdef_field_metadata(
        comp_name,
        grid_name,
        "clock",
        "{\n"
        "  \"cf\": {\n"
        "    \"standard_name\": \"t\",\n"
        "    \"long_name\": \"time\",\n"
        "    \"units\": \"s\"\n"
        "  }\n"
        "}");

    yac_cenddef();

    size_t datasize    = nbr_points[0] * nbr_points[1];
    double *x_data     = malloc(datasize * sizeof(*x_data)),
           *y_data     = malloc(datasize * sizeof(*y_data)),
           *z_data     = malloc(datasize * sizeof(*z_data)),
           *clock_data = malloc(datasize * sizeof(*clock_data));
    double dx          = M_PI / resolution;

    // this seems to be the mapping lon,lat -> xyz healpix uses
    for (size_t i = 0; i < nbr_points[1]; ++i) {
        for (size_t j = 0; j < nbr_points[0]; ++j) {
            size_t idx      = j + nbr_points[0] * i;
            double lon      = j * dx;
            double lat      = i * dx - M_PI_2;
            x_data[idx]     = cos(lon) * cos(lat);
            y_data[idx]     = sin(lon) * cos(lat);
            z_data[idx]     = sin(lat);
            clock_data[idx] = 0;
        }
    }

    int info = 0, ierror = -1;
    int step = 0;
    while (info != YAC_ACTION_PUT_FOR_RESTART) {
        info = YAC_ACTION_PUT_FOR_RESTART;
        printf("step: %d", step);
        if (yac_cget_role_from_field_id(x_field_id) != YAC_EXCHANGE_TYPE_NONE) {
            printf(" time: %s\n", yac_cget_field_datetime(x_field_id));
            yac_cput_(x_field_id, 1, x_data, &info, &ierror);
        }
        if (yac_cget_role_from_field_id(y_field_id) != YAC_EXCHANGE_TYPE_NONE)
            yac_cput_(y_field_id, 1, y_data, &info, &ierror);
        if (yac_cget_role_from_field_id(z_field_id) != YAC_EXCHANGE_TYPE_NONE)
            yac_cput_(z_field_id, 1, z_data, &info, &ierror);
        if (yac_cget_role_from_field_id(clock_field_id) != YAC_EXCHANGE_TYPE_NONE)
            yac_cput_(clock_field_id, 1, clock_data, &info, &ierror);
        for (size_t i = 0; i < datasize; ++i)
            clock_data[i] += 1;
        printf("\n");
        ++step;
    }

    yac_cfinalize();

    free(x_data);
    free(y_data);
    free(z_data);
    free(clock_data);
}
